@@ -121,16 +121,18 @@ class CloudFlare(object):
             """ Cloudflare v4 API"""
 
             if self.logger:
-                self.logger.debug('Call: %s,%s,%s,%s,%s,%s' % (str(parts[0]),
-                                                               str(identifier1),
-                                                               str(parts[1]),
-                                                               str(identifier2),
-                                                               str(parts[2]),
-                                                               str(identifier3)))
-                self.logger.debug('Call: optional params and data %s %s' % (str(params),
-                                                                            str(data)))
+                self.logger.debug('Call: %s,%s,%s,%s,%s,%s',
+                                  str(parts[0]),
+                                  str(identifier1),
+                                  str(parts[1]),
+                                  str(identifier2),
+                                  str(parts[2]),
+                                  str(identifier3))
+                self.logger.debug('Call: optional params and data %s %s',
+                                  str(params),
+                                  str(data))
                 if files:
-                    self.logger.debug('Call: upload file %r' % (files))
+                    self.logger.debug('Call: upload file %r', files)
 
             if (method is None) or (parts[0] is None):
                 # should never happen
@@ -164,8 +166,8 @@ class CloudFlare(object):
                 url += '/' + identifier3
 
             if self.logger:
-                self.logger.debug('Call: method and url %s %s' % (str(method), str(url)))
-                self.logger.debug('Call: headers %s' % str(sanitize_secrets(headers)))
+                self.logger.debug('Call: method and url %s %s', str(method), str(url))
+                self.logger.debug('Call: headers %s', str(sanitize_secrets(headers)))
 
             method = method.upper()
 
@@ -259,7 +261,8 @@ class CloudFlare(object):
                 response_data = response_data.decode("utf-8")
 
             if self.logger:
-                self.logger.debug('Response: %d, %s, %s' % (response_code, response_type, response_data))
+                self.logger.debug('Response: %d, %s, %s',
+                                  response_code, response_type, response_data)
 
             if response_code >= 500 and response_code <= 599:
                 # 500 Internal Server Error
@@ -310,29 +313,43 @@ class CloudFlare(object):
 
             [response_type, response_code, response_data] = self._network(method,
                                                                           headers, parts,
-                                                                          identifier1, identifier2, identifier3,
+                                                                          identifier1,
+                                                                          identifier2,
+                                                                          identifier3,
                                                                           params, data, files)
 
             if response_type == 'application/json':
                 # API says it's JSON; so it better be parsable as JSON
+                # NDJSON is returned by Enterprise Log Share i.e. /zones/:id/logs/received
+                if hasattr(response_data, 'decode'):
+                    response_data = response_data.decode('utf-8')
                 try:
-                    if hasattr(response_data, 'decode'):
-                        response_data = json.loads(response_data.decode('utf-8'))
-                    else:
-                        response_data = json.loads(response_data)
+                    response_data = json.loads(response_data)
                 except ValueError:
                     if response_data == '':
                         # This should really be 'null' but it isn't. Even then, it's wrong!
                         if response_code == requests.codes.ok:
                             # 200 ok
-                            response_data = {'success': True, 'result': None}
+                            response_data = {'success': True,
+                                             'result': None}
                         else:
                             # 3xx & 4xx errors
-                            response_data = {'success': False, 'code': response_code, 'result': None}
+                            response_data = {'success': False,
+                                             'code': response_code,
+                                             'result': None}
                     else:
-                        # While this should not happen; it's always possible
-                        self.logger.debug('Response data not JSON: %r' % (response_data))
-                        raise CloudFlareAPIError(0, 'JSON parse failed - report to Cloudflare.')
+                        # Lets see if it's NDJSON data
+                        # NDJSON is a series of JSON elements with newlines between each element
+                        try:
+                            r = []
+                            for l in response_data.splitlines():
+                                r.append(json.loads(l))
+                            response_data = r
+                        except:
+                            # While this should not happen; it's always possible
+                            if self.logger:
+                                self.logger.debug('Response data not JSON: %r', response_data)
+                            raise CloudFlareAPIError(0, 'JSON parse failed - report to Cloudflare.')
 
                 if response_code == requests.codes.ok:
                     # 200 ok - so nothing needs to be done
@@ -343,11 +360,10 @@ class CloudFlare(object):
                     pass
             elif response_type == 'text/plain' or response_type == 'application/octet-stream':
                 # API says it's text; but maybe it's actually JSON? - should be fixed in API
+                if hasattr(response_data, 'decode'):
+                    response_data = response_data.decode('utf-8')
                 try:
-                    if hasattr(response_data, 'decode'):
-                        response_data = json.loads(response_data.decode('utf-8'))
-                    else:
-                        response_data = json.loads(response_data)
+                    response_data = json.loads(response_data)
                 except ValueError:
                     # So it wasn't JSON - moving on as if it's text!
                     # A single value is returned (vs an array or object)
@@ -356,24 +372,44 @@ class CloudFlare(object):
                         response_data = {'success': True, 'result': str(response_data)}
                     else:
                         # 3xx & 4xx errors
-                        response_data = {'success': False, 'code': response_code, 'result': str(response_data)}
+                        response_data = {'success': False,
+                                         'code': response_code,
+                                         'result': str(response_data)}
             elif response_type == 'text/javascript' or response_type == 'application/javascript':
                 # used by Cloudflare workers
                 if response_code == requests.codes.ok:
                     # 200 ok
-                    response_data = {'success': True, 'result': str(response_data)}
+                    response_data = {'success': True,
+                                     'result': str(response_data)}
                 else:
                     # 3xx & 4xx errors
-                    response_data = {'success': False, 'code': response_code, 'result': str(response_data)}
+                    response_data = {'success': False,
+                                     'code': response_code,
+                                     'result': str(response_data)}
+            elif response_type == 'text/html':
+                # used by media for preview
+                if response_code == requests.codes.ok:
+                    # 200 ok
+                    response_data = {'success': True,
+                                     'result': str(response_data)}
+                else:
+                    # 3xx & 4xx errors
+                    response_data = {'success': False,
+                                     'code': response_code,
+                                     'result': str(response_data)}
+
             else:
                 # Assuming nothing - but continuing anyway
                 # A single value is returned (vs an array or object)
                 if response_code == requests.codes.ok:
                     # 200 ok
-                    response_data = {'success': True, 'result': str(response_data)}
+                    response_data = {'success': True,
+                                     'result': str(response_data)}
                 else:
                     # 3xx & 4xx errors
-                    response_data = {'success': False, 'code': response_code, 'result': str(response_data)}
+                    response_data = {'success': False,
+                                     'code': response_code,
+                                     'result': str(response_data)}
 
             # it would be nice to return the error code and content type values; but not quite yet
             return response_data
@@ -417,22 +453,25 @@ class CloudFlare(object):
                     message = errors['error']
                 else:
                     message = ''
+                if 'messages' in response_data:
+                    errors['error_chain'] = response_data['messages']
                 if 'error_chain' in errors:
                     error_chain = errors['error_chain']
                     for error in error_chain:
                         if self.logger:
-                            self.logger.debug('Response: error %d %s - chain' %
-                                              (error['code'], error['message']))
+                            self.logger.debug('Response: error %d %s - chain',
+                                              error['code'],
+                                              error['message'])
                     if self.logger:
-                        self.logger.debug('Response: error %d %s' % (code, message))
+                        self.logger.debug('Response: error %d %s', code, message)
                     raise CloudFlareAPIError(code, message, error_chain)
                 else:
                     if self.logger:
-                        self.logger.debug('Response: error %d %s' % (code, message))
+                        self.logger.debug('Response: error %d %s', code, message)
                     raise CloudFlareAPIError(code, message)
 
             if self.logger:
-                self.logger.debug('Response: %s' % (response_data['result']))
+                self.logger.debug('Response: %s', response_data['result'])
             if self.raw:
                 result = {}
                 # theres always a result value
@@ -455,18 +494,18 @@ class CloudFlare(object):
                                       identifier1, identifier2, identifier3,
                                       params, data, files)
             if self.logger:
-                self.logger.debug('Response: %s' % (response_data))
+                self.logger.debug('Response: %s', response_data)
             result = response_data
             return result
 
     class _add_unused(object):
         """ Cloudflare v4 API"""
 
-        def __init__(self, base, api_call_part1, api_call_part2=None, api_call_part3=None):
+        def __init__(self, base, p1, p2=None, p3=None):
             """ Cloudflare v4 API"""
 
             self._base = base
-            self._parts_unused = [api_call_part1, api_call_part2, api_call_part3]
+            self._parts_unused = [p1, p2, p3]
 
         def __call__(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
             """ Cloudflare v4 API"""
@@ -489,7 +528,7 @@ class CloudFlare(object):
 
             raise CloudFlareAPIError(0, 'patch() call not available for this endpoint')
 
-        def post(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
+        def post(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None, files=None):
             """ Cloudflare v4 API"""
 
             raise CloudFlareAPIError(0, 'post() call not available for this endpoint')
@@ -507,11 +546,11 @@ class CloudFlare(object):
     class _add_noauth(object):
         """ Cloudflare v4 API"""
 
-        def __init__(self, base, api_call_part1, api_call_part2=None, api_call_part3=None):
+        def __init__(self, base, p1, p2=None, p3=None):
             """ Cloudflare v4 API"""
 
             self._base = base
-            self._parts = [api_call_part1, api_call_part2, api_call_part3]
+            self._parts = [p1, p2, p3]
 
         def __call__(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
             """ Cloudflare v4 API"""
@@ -536,7 +575,7 @@ class CloudFlare(object):
 
             raise CloudFlareAPIError(0, 'patch() call not available for this endpoint')
 
-        def post(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
+        def post(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None, files=None):
             """ Cloudflare v4 API"""
 
             raise CloudFlareAPIError(0, 'post() call not available for this endpoint')
@@ -554,11 +593,11 @@ class CloudFlare(object):
     class _add_with_auth(object):
         """ Cloudflare v4 API"""
 
-        def __init__(self, base, api_call_part1, api_call_part2=None, api_call_part3=None):
+        def __init__(self, base, p1, p2=None, p3=None):
             """ Cloudflare v4 API"""
 
             self._base = base
-            self._parts = [api_call_part1, api_call_part2, api_call_part3]
+            self._parts = [p1, p2, p3]
 
         def __call__(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
             """ Cloudflare v4 API"""
@@ -609,11 +648,11 @@ class CloudFlare(object):
     class _add_with_auth_unwrapped(object):
         """ Cloudflare v4 API"""
 
-        def __init__(self, base, api_call_part1, api_call_part2=None, api_call_part3=None):
+        def __init__(self, base, p1, p2=None, p3=None):
             """ Cloudflare v4 API"""
 
             self._base = base
-            self._parts = [api_call_part1, api_call_part2, api_call_part3]
+            self._parts = [p1, p2, p3]
 
         def __call__(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
             """ Cloudflare v4 API"""
@@ -664,11 +703,11 @@ class CloudFlare(object):
     class _add_with_cert_auth(object):
         """ Cloudflare v4 API"""
 
-        def __init__(self, base, api_call_part1, api_call_part2=None, api_call_part3=None):
+        def __init__(self, base, p1, p2=None, p3=None):
             """ Cloudflare v4 API"""
 
             self._base = base
-            self._parts = [api_call_part1, api_call_part2, api_call_part3]
+            self._parts = [p1, p2, p3]
 
         def __call__(self, identifier1=None, identifier2=None, identifier3=None, params=None, data=None):
             """ Cloudflare v4 API"""
@@ -716,6 +755,42 @@ class CloudFlare(object):
                                                  identifier1, identifier2, identifier3,
                                                  params, data)
 
+    def add(self, t, p1, p2=None, p3=None):
+        """add api call to class"""
+
+        a = []
+        if p1:
+            a += p1.split('/')
+        if p2:
+            a += p2.split('/')
+        if p3:
+            a += p3.split('/')
+
+        branch = self
+        for element in a[0:-1]:
+            try:
+                branch = getattr(branch, element)
+            except:
+                # should never happen
+                raise CloudFlareAPIError(0, 'api load name failed')
+        name = a[-1]
+
+        if t == 'VOID':
+            f = self._add_unused(self._base, p1, p2, p3)
+        elif t == 'OPEN':
+            f = self._add_noauth(self._base, p1, p2, p3)
+        elif t == 'AUTH':
+            f = self._add_with_auth(self._base, p1, p2, p3)
+        elif t == 'CERT':
+            f = self._add_with_cert_auth(self._base, p1, p2, p3)
+        elif t == 'AUTH_UNWRAPPED':
+            f = self._add_with_auth_unwrapped(self._base, p1, p2, p3)
+        else:
+            # should never happen
+            raise CloudFlareAPIError(0, 'api load type mismatch')
+
+        setattr(branch, name, f)
+
     def api_list(self, m=None, s=''):
         """recursive walk of the api tree returning a list of api calls"""
         if m is None:
